* Program: SRCPARSE.PRG
*  Action: Creates databases containing one source line per record
*  Syntax: DO SRCPARSE WITH <expC>, <expC>, [<expC>]
*  Notice: Copyright (c) 1994 Solid Rock Systemsm, All Rights Reserved
*        : parse_prg (c) 1993 Rettig Micro Corp. All Rights Reserved
* Version: 2.0
*  Update: March 21, 1994
*        : April 1, 1994
*  Syntax: DO SRCPARSE WITH <srcFile>, <lineDBF>, <xrefDBF>
*   Notes: srcFile - the name of the source file to parse.  It can be a
*        :             PRG, MPR, or SPR
*        : lineDBF - the name of the source line DBF.  This database
*        :             contains the following fields:
*        :               Module      C(12)   Source file name
*        :               FcnName     C(10)   Function name within Module
*        :               ModuleLine  N(6,0)  Line # within entire module
*        :               FcnLine     N(6,0)  Line # within function
*        :               Source      C(80)   Source code text
*        : xrefDBF - the name of the xref DBF.  This database contains 
*        :             the following fields:
*        :               Module      C(12)   Source file name
*        :               FcnName     C(10)   Function name within Module
*        :               ModuleLine  N(6,0)  Line # within entire module
*        :               FirstLine   N(6,0)  Continuation line's 1st line
*        :             This database is a cross-reference of the actual
*        :             module line # to the first line # for that
*        :             continuation set.  IOW, for single-line source 
*        :             lines, ModuleLine = FirstLine.  For each line of a
*        :             multi-line set, FirstLine is the ModuleLine that
*        :             contains the first line of that set.
*        : Either of the two databases will be created if they don't
*        :    exist.  If they do exist, any existing records for srcFile
*        :    will be deleted before processing.
*        : If you can live with seeing the *last* line of a multi-line
*        :    set in your trace instead of the first, change the __XREF
*        :    #define below to a .F.  This will remove all of the
*        :    references to the XREF database.
*
* Preprocessor directives.
#DEFINE __XREF         .T.
#DEFINE cNOFILENAME    "Must specify a source file"
#DEFINE cBADFILENAME   "Specified source file does not exist"
#DEFINE cDIRDELIM      "\"
#DEFINE cEXTDELIM      "."
#DEFINE cPRG           ".PRG"
#DEFINE cCRLF          CHR(13)+CHR(10)
#DEFINE cREADBYTES     64000

* Begin code.
PARAMETERS p_filename, p_linedbf, p_xrefdbf
PRIVATE p_filename, p_linedbf, p_xrefdbf, f_filename, ;
        llExclusive, llResource, llSafety, llTalk

* If no xrefDBF, initialize it to empty
#IF __XREF
#ELSE
   p_xrefDBF = ""
#ENDIF
   
* Set up our environment.
llTalk      = SET('TALK') == "ON"
llSafety    = SET('SAFETY') == "ON"
llExclusive = SET('EXCLUSIVE') == "ON"
llResource  = SET('RESOURCE') == "ON"
SET TALK OFF
SET EXCLUSIVE ON
SET RESOURCE OFF

IF EMPTY(m.p_filename)
   ? cNOFILENAME
   RETURN
ELSE
   f_fullname = FULLPATH(m.p_filename)
   f_filename = SUBSTR(m.f_fullname, RAT(cDIRDELIM,m.f_fullname)+1)
   f_ext      = SUBSTR(f_filename, RAT(cEXTDELIM, f_filename))
   IF EMPTY(f_ext)
      f_fullname = f_fullname + cPRG
      f_filename = f_filename + cPRG
   ENDIF
ENDIF

* If we can't find the specified source file, display error and exit
IF NOT FILE(m.f_fullname)
   ? cBADFILENAME + " " + m.f_fullname
   RETURN  && early exit
ENDIF

* Create the temporary source cursor
SET SAFETY OFF
CREATE CURSOR Src ;
   (Module C(12), FcnName C(10), Position N(10,0), Size N(6,0), ;
    Detail M(10))

IF llSafety
   SET SAFETY ON
ENDIF

* Open specified program file, parse into table
f_srcHandle = FOPEN(m.f_fullname)
DO parse_prg WITH m.f_srcHandle, UPPER(m.f_filename)
=FCLOSE(m.f_srcHandle)

* Parse memo into individual records
DO parse_lines WITH UPPER(m.p_linedbf), UPPER(m.p_xrefdbf)
USE IN src

* Be nice and put the environment like we found it
IF llExclusive
   SET EXCLUSIVE ON
ENDIF

IF llTalk
   SET TALK ON
ENDIF

IF llResource
   SET RESOURCE ON
ENDIF

RETURN


*************************************************************
***                  SRCPARSE FUNCTIONS                   ***
*************************************************************

FUNCTION waitMsg
PARAMETERS p_msg
IF _WINDOWS OR _MAC
   SET MESSAGE TO p_msg
ELSE
   WAIT WINDOW NOWAIT p_msg
ENDIF
RETURN
*** waitMsg ************************************************




FUNCTION parse_prg
* Parses a source file into a table with a record for each FUNCTION.
PARAMETERS p_handle, p_fname
PRIVATE    p_handle, p_fname
#define cOFFSET   100

PRIVATE f_filename, f_ext, f_fname, f_pos, f_text, f_i, f_j
DIMENSION laExt[10,2]

* Define our extension cross-reference:
*  [x,1] is the source extension
*  [x,2] is the compiled source extension (returned by SYS(16))
laExt[1,1] = ".PRG"
laExt[1,2] = ".FXP"
laExt[2,1] = ".SPR"
laExt[2,2] = ".SPX"
laExt[3,1] = ".MPR"
laExt[4,1] = ".MPX"

* Split the source filename into its file and extension components
f_filename = LEFT(p_fname, RAT(cEXTDELIM, p_fname)-1)
f_ext      = SUBSTR(p_fname, RAT(cEXTDELIM, p_fname))
* Determine the correct compiled extension ...
f_pos      = ASCAN(laExt, m.f_ext)
IF m.f_pos == 0
	DO waitMsg WITH cBADEXT + " " + f_filename
	RETURN
ELSE
	f_fname = m.f_filename + laExt[m.f_pos+1]
ENDIF

* Start off with the module name as the function name
INSERT INTO Src VALUES (m.f_fname, m.f_filename, 0, 0, "")

=FSEEK(m.p_handle,0)  && point to file start
FOR f_i=1 TO 1000  && freads per file
   f_text = FREAD(m.p_handle, cREADBYTES)
   FOR f_j=1 TO 1000  && functions in fread section of file
      f_pos = AT(cCRLF+'FUNCTION ', m.f_text, m.f_j)
      IF m.f_pos==0 OR AT(cCRLF, SUBSTR(m.f_text, m.f_pos+2, 255))==0
         EXIT
      ELSE
         INSERT INTO Src VALUES;
            (m.f_fname, UPPER(SUBSTR(m.f_text, m.f_pos+11,;
                          AT(cCRLF, SUBSTR(m.f_text,;
                                           m.f_pos+11, 255))-1)),;
             m.f_pos+2 + ((cREADBYTES-cOFFSET)*(m.f_i-1)),;
             0,;
             "")
         DO waitMsg WITH "Reading: " + RTRIM(m.p_fname) +  ;
                            ", Function " + src.FcnName
         SKIP -1
         IF NOT (m.f_pos=0 OR BOF())
            =FSEEK(m.p_handle, IIF(src.Position==0,;
                                   0, src.Position-1))
            REPLACE src.Size WITH m.f_pos+2 +;
               ((cREADBYTES-cOFFSET)*(m.f_i-1)) -;
                  IIF(src.Position==0, 1, src.Position),;
               src.Detail WITH FREAD(m.p_handle, src.Size)
         ENDIF
         SKIP
      ENDIF
   ENDFOR
   IF FEOF(m.p_handle)
      f_pos = FSEEK(m.p_handle, 0, 2)
      INSERT INTO Src VALUES;
         (m.f_fname,;
			 "EOF",;
          m.f_pos,;
          0,;
          "")
      SKIP -1
      =FSEEK(m.p_handle, IIF(src.Position==0,;
                             0, src.Position-1))
      REPLACE src.Size;
         WITH m.f_pos - IIF(src.Position==0,;
                            0, src.Position-1),;
              src.Detail WITH FREAD(m.p_handle, src.Size)
      EXIT
   ELSE
      * Back up cOFFSET bytes in case a FUNCTION line splits
      * between file reads.
      =FSEEK(m.p_handle, (cREADBYTES-cOFFSET)*m.f_i)
   ENDIF
ENDFOR

RETURN
*** parse_prg ***********************************************



FUNCTION parse_lines
PARAMETER p_lineDBF, p_xrefDBF
PRIVATE f_lineDBF, f_xrefDBF, lnExtPos, lnDirPos, lnMemoWidth, ;
        lnModuleLine, lnFcnLine, lnNumLines, lcSourceLn, lnFirstLine, ;
        llDeleted, llContLine, llIsCont, lnSeconds

#DEFINE cMAXWIDTH 255
#DEFINE cCONTCHAR ";"
#DEFINE cDBFEXT ".DBF"
#DEFINE cNOLINEDBF "Must specify a source line DBF name"
#DEFINE cNOXREFDBF "Must specify a source line cross-reference DBF name"

* Make sure a line DBF was specified, and that it includes an extension
IF EMPTY(m.p_lineDBF)
   DO waitMsg WITH cNOLINEDBF
   RETURN
ELSE
   lnExtPos = RAT(cEXTDELIM,m.p_lineDBF)
   lnDirPos = RAT(cDIRDELIM,m.p_lineDBF)
   IF (m.lnExtPos > m.lnDirPos)
      f_lineDBF = LEFT(m.p_lineDBF, m.lnExtPos-1)
   ELSE
      f_lineDBF = m.p_lineDBF
   ENDIF
ENDIF

#IF __XREF
* Make sure a xref DBF was specified, and that it includes an extension
IF EMPTY(m.p_xrefDBF)
   DO waitMsg WITH cNOXREFDBF
   RETURN
ELSE
   lnExtPos = RAT(cEXTDELIM,m.p_xrefDBF)
   lnDirPos = RAT(cDIRDELIM,m.p_xrefDBF)
   IF (m.lnExtPos > m.lnDirPos)
      f_xrefDBF = LEFT(p_xrefDBF, m.lnExtPos-1)
   ELSE
      f_xrefDBF = m.p_xrefDBF
   ENDIF
ENDIF
#ENDIF

lnMemoWidth = SET('MEMOWIDTH')
llDeleted   = SET('DELETED') == "ON"
SET MEMOWIDTH TO cMAXWIDTH
SET DELETED OFF

DO waitMsg WITH "Cleaning up DBF's"
SELECT 0
* If the line DBF doesn't exist, create it
IF NOT FILE(f_lineDBF + cDBFEXT)
   CREATE TABLE (f_lineDBF);
      (Module C(12), FcnName C(12), ModuleLine N(6,0), FcnLine N(6,0), ;
		 Source C(80))
   USE (f_lineDBF)
ELSE
   * If it does, clean out any existing records for this module
   USE (f_lineDBF)
   DELETE TAG ALL
   DELETE FOR Module = src.module
   COUNT FOR DELETED() TO m.lnDeleted
   IF m.lnDeleted > 0
      PACK
   ENDIF
ENDIF

#IF __XREF
SELECT 0
* If the xref DBF doesn't exist, create it
IF NOT FILE(f_xrefDBF + cDBFEXT)
	CREATE TABLE (f_xrefDBF);
		(Module C(12), FcnName C(12), ModuleLine N(6,0), FirstLine N(6,0))
    USE (f_xrefDBF)
ELSE
   * If it does, clean out any existing records for this module
   USE (f_xrefDBF)
   DELETE TAG ALL
   DELETE FOR Module = src.module
   COUNT FOR DELETED() TO lnDeleted
   IF m.lnDeleted > 0
      PACK
   ENDIF
ENDIF
USE											&& Close the file
#ENDIF

lnSeconds = SECONDS()
SELECT Src							        && Cursor created by main module
lnModuleLine = 1
SCAN
   SCATTER MEMO MEMVAR
   DO waitMsg WITH "Parsing: Module " + RTRIM(Src.Module) + ;
                      ", Function " + Src.FcnName
   _MLINE = 0
   * DON'T touch the following line.  For some reason, MEMLINES returns
   * one too many lines.  If you look in the Src cursor, there appears
   * to be one more blank line at the end of each function.  *However*,
   * if you take the the cursor and append all of the memo's together,
   * it produces a source file that *exactly* matches the original.
   * Don't know why, don't *want* to know why <g>.
   lnNumLines = IIF(RECNO()!=RECCOUNT(),MEMLINES(detail)-1,MEMLINES(detail))
   llContLine = .F.
   FOR m.lnFcnLine = 1 TO lnNumLines
      lcSourceLn = MLINE(detail, 1, _MLINE)
      INSERT INTO (f_lineDBF) ;
         VALUES (m.Module, m.FcnName, m.lnModuleLine, m.lnFcnLine, ;
					  m.lcSourceLn)

#IF __XREF
      * 1st Line is left alone if we've been in a continuation line
      lnFirstLine  = IIF(llContLine, m.lnFirstLine, m.lnModuleLine)
      INSERT INTO (f_xrefDBF) ;
         VALUES (m.Module, m.FcnName, m.lnModuleLine, m.lnFirstLine)

      * Is the current line a continuation line?
      llIsCont     = RIGHT(RTRIM(lcSourceLn),1) == cCONTCHAR
      * Our continuation status is now the status of *this* line
      llContLine   = m.llIsCont
#ENDIF
      * Increment the module's line count
      lnModuleLine = m.lnModuleLine + 1
   ENDFOR
ENDSCAN

DO waitMsg WITH "Parsing took " + ;
                   LTRIM(STR(SECONDS()-m.lnSeconds, 7, 3)) + " seconds"

SELECT (f_lineDBF)
INDEX ON ModuleLine TAG ModuleLine
INDEX ON Module     TAG Module
USE

#IF __XREF
SELECT (f_xrefDBF)
INDEX ON ModuleLine TAG ModuleLine
INDEX ON Module     TAG Module
USE
#ENDIF

SET MEMOWIDTH TO (lnMemoWidth)
IF llDeleted
   SET DELETED ON
ENDIF

RETURN
*** parse_lines ***********************************************

*** SRCparse.prg **********************************************